import os
import json
from collections import OrderedDict
import itertools
import requests

import django_rq
import pysolr

from django.core.management.base import BaseCommand, CommandError
from django.core.mail import EmailMultiAlternatives

from django.db import transaction, connection
from django.db.utils import ProgrammingError

from django.template.loader import get_template
from django.conf import settings

from django.contrib.auth.models import User

try:
    haystack_url = settings.HAYSTACK_CONNECTIONS['default']['URL']
except KeyError:
    haystack_url = None

class Command(BaseCommand):
    help = 'Send email notifications to subscribed users'
    
    def add_arguments(self, parser):
        parser.add_argument(
            '--users',
            default='all',
            help='Comma separated list of usernames to send notifications to.'
        )
    
    def handle(self, *args, **options):
        
        subscribed_users = ''' 
            SELECT 
              u.id AS user_id,
              MAX(u.email) as user_email,
              array_agg(DISTINCT bas.bill_id) AS bill_action_ids,
              array_agg(DISTINCT bss.search_params) AS bill_search_params,
              array_agg(DISTINCT cas.committee_id) AS committee_action_ids,
              array_agg(DISTINCT ces.committee_id) AS committee_event_ids,
              array_agg(DISTINCT ps.person_id) as person_ids
            FROM auth_user AS u
            LEFT JOIN notifications_billactionsubscription AS bas
              ON u.id = bas.user_id
            LEFT JOIN notifications_billsearchsubscription AS bss
              ON u.id = bss.user_id
            LEFT JOIN notifications_committeeactionsubscription AS cas
              ON u.id = cas.user_id
            LEFT JOIN notifications_committeeeventsubscription AS ces
              ON u.id = ces.user_id
            LEFT JOIN notifications_personsubscription AS ps
              ON u.id = ps.user_id
            WHERE (bas.bill_id IS NOT NULL
                   OR bss.search_params IS NOT NULL
                   OR cas.committee_id IS NOT NULL
                   OR ces.committee_id IS NOT NULL
                   OR ps.person_id IS NOT NULL)
        '''
        
        q_args = []

        if options['users'] != 'all':
            users = tuple(options['users'].split(','))
            subscribed_users = '{} AND u.username IN %s'.format(subscribed_users)
            q_args.append(users)
            

        subscribed_users = '{} GROUP BY u.id'.format(subscribed_users)
        cursor = connection.cursor()
        cursor.execute(subscribed_users, q_args)
        columns = [c[0] for c in cursor.description]
        
        for row in cursor:
            user_subscriptions = dict(zip(columns, row))
            
            bill_action_updates = []
            bill_search_updates = []
            committee_action_updates = []
            committee_event_updates = []
            person_updates = []
            
            bill_action_ids = [i for i in user_subscriptions['bill_action_ids'] if i]
            bill_search_params = [i for i in user_subscriptions['bill_search_params'] if i]
            committee_action_ids = [i for i in user_subscriptions['committee_action_ids'] if i]
            committee_event_ids = [i for i in user_subscriptions['committee_event_ids'] if i]
            person_ids = [i for i in user_subscriptions['person_ids'] if i]
            
            send_notification = False

            if bill_action_ids:
                bill_action_updates = self.find_bill_action_updates(bill_action_ids)
                
                if bill_action_updates:
                    send_notification = True

            if bill_search_params:
                bill_search_updates = self.find_bill_search_updates(bill_search_params)

                if bill_search_updates:
                    send_notification = True

            if committee_action_ids:
                committee_action_updates = self.find_committee_action_updates(committee_action_ids)

                if committee_action_updates:
                    send_notification = True

            if committee_event_ids:
                committee_event_updates = self.find_committee_event_updates(committee_event_ids)

                if committee_event_updates:
                    send_notification = True
            
            if person_ids:
                person_updates = self.find_person_updates(person_ids)

                if person_updates:
                    send_notification = True
            
            if send_notification:
                send_notification_email.delay(user_id=user_subscriptions['user_id'],
                                              user_email=user_subscriptions['user_email'],
                                              bill_action_updates=bill_action_updates,
                                              bill_search_updates=bill_search_updates,
                                              person_updates=person_updates,
                                              committee_action_updates=committee_action_updates,
                                              committee_event_updates=committee_event_updates)

    def find_bill_action_updates(self, bill_ids):
        
        new_actions = ''' 
            SELECT DISTINCT ON (bill.ocd_id)
              bill.slug AS bill_slug,
              bill.identifier AS bill_identifier,
              bill.description AS bill_description,
              action.description AS action_description,
              action.date AS action_date
            FROM new_action AS new
            JOIN councilmatic_core_bill AS bill
              ON new.bill_id = bill.ocd_id
            JOIN councilmatic_core_action AS action
              ON new.bill_id = action.bill_id
            WHERE new.bill_id IN %s
        '''
        
        cursor = connection.cursor()
        cursor.execute(new_actions, [tuple(bill_ids)])
        
        bill_action_updates = []

        for row in cursor:
            
            bill = {
                'slug': row[0],
                'identifier': row[1],
                'description': row[2],
            }
            action = {
                'description': row[3],
                'date': row[4],
            }

            bill_action_updates.append((bill, action))
            
        return bill_action_updates

    def find_bill_search_updates(self, search_params):
        
        if not haystack_url:
            self.stdout.write(self.style.ERROR('Solr is not configured so no search notifications will be sent'))
            return []
        
        search_updates = []
        
        cursor = connection.cursor()

        for params in search_params:
            query_params = {
                'q': params['term'], 
                'fq': [],
                'wt': 'json'
            }

            for facet, values in params['facets'].items():
                for value in values:
                    query_params['fq'].append('{0}:{1}'.format(facet, value))
            
            results = requests.get('{}/select'.format(haystack_url), params=query_params)
            
            ocd_ids = tuple(r['ocd_id'] for r in results.json()['response']['docs'])
            
            new_bills = ''' 
                SELECT DISTINCT ON (bill.ocd_id)
                  bill.slug AS bill_slug,
                  bill.identifier AS bill_identifier,
                  bill.description AS bill_description
                FROM new_bill AS new
                JOIN councilmatic_core_bill AS bill
                  ON new.ocd_id = bill.ocd_id
                WHERE new.ocd_id IN %s
            '''
            
            cursor.execute(new_bills, [ocd_ids])
            
            bills = []

            for row in cursor:
                bill = {
                    'slug': row[0],
                    'identifier': row[1],
                    'description': row[2]
                }
                bills.append(bill)

            if bills:
                search_updates.append({
                    'params': params,
                    'bills': bills
                })
        
        return search_updates
    
    def find_person_updates(self, person_ids):
        # If person sponsors something new
        # If bill sponsored by person has new or updated action
        
        new_sponsorships = ''' 
            SELECT DISTINCT ON (person.ocd_id, bill.ocd_id)
              person.name,
              person.slug,
              bill.identifier,
              bill.slug,
              bill.description
            FROM new_sponsorship AS new
            JOIN councilmatic_core_person AS person
              ON new.person_id = person.ocd_id
            JOIN councilmatic_core_bill AS bill
              ON new.bill_id = bill.ocd_id
            WHERE new.person_id IN %s
        '''

        new_actions = ''' 
            SELECT DISTINCT ON (person.ocd_id, bill.ocd_id)
              person.name,
              person.slug,
              bill.identifier,
              bill.slug, 
              bill.description,
              action.description,
              action.date
            FROM new_action AS new
            JOIN councilmatic_core_bill AS bill
              ON new.bill_id = bill.ocd_id
            JOIN councilmatic_core_action AS action
              ON bill.ocd_id = action.bill_id
            JOIN councilmatic_core_sponsorship AS sponsor
              ON bill.ocd_id = sponsor.bill_id
            JOIN councilmatic_core_person AS person
              ON sponsor.person_id = person.ocd_id
            WHERE person.ocd_id IN %s
        '''
        
        person_updates = []
        
        cursor = connection.cursor()
        cursor.execute(new_sponsorships, [tuple(person_ids)])

        for row in cursor:
            person = {
                'name': row[0],
                'slug': row[1],
            }
            bill = {
                'identifier': row[2],
                'slug': row[3],
                'description': row[4],
                'action_description': None,
                'action_date': None,
                'update_type': 'New Sponsorship'
            }
            person_updates.append((person, bill))

        cursor.execute(new_actions, [tuple(person_ids)])

        for row in cursor:
            person = {
                'name': row[0],
                'slug': row[1],
            }
            bill = {
                'identifier': row[2],
                'slug': row[3],
                'description': row[4],
                'action_description': row[5],
                'action_date': row[6],
                'update_type': 'New Action'
            }
            person_updates.append((person, bill))
        
        update_groups = []

        outer_grouper = lambda x: x[0]['slug']
        person_updates = sorted(person_updates, key=outer_grouper)
        
        inner_grouper = lambda x: x[1]['update_type']

        for slug, group in itertools.groupby(person_updates, key=outer_grouper):
            bill_group = {}
            
            group = sorted(group, key=inner_grouper)

            for update_type, inner_group in itertools.groupby(group, key=inner_grouper):
                bill_group[update_type] = {}
                
                for person, bill in inner_group:
                    bill_group[update_type].update(person)

                    try:
                        bill_group[update_type]['bills'].append(bill)
                    except KeyError:
                        bill_group[update_type]['bills'] = [bill]
            
            update_groups.append(bill_group)
        
        return update_groups
    
    def find_committee_action_updates(self, committee_ids):
        # New actions taken by a committee on a bill

        new_actions = ''' 
            SELECT DISTINCT ON (committee.ocd_id, bill.ocd_id, action.id)
              committee.name,
              committee.slug,
              bill.identifier,
              bill.slug,
              bill.description,
              action.description,
              action.date,
              action.order
            FROM councilmatic_core_organization AS committee
            JOIN councilmatic_core_action AS action
              ON committee.ocd_id = action.organization_id
            JOIN councilmatic_core_bill AS bill 
              ON action.bill_id = bill.ocd_id
            JOIN new_action AS new
              ON bill.ocd_id = new.bill_id
            WHERE committee.ocd_id IN %s
            ORDER BY committee.ocd_id,
                     bill.ocd_id,
                     action.id,
                     action.order DESC
        '''
        
        cursor = connection.cursor()
        cursor.execute(new_actions, [tuple(committee_ids)])
        
        committee_updates = []
        
        outer_grouper = lambda x: x[1]
        inner_grouper = lambda x: x[3]
        groups = sorted(cursor, key=outer_grouper)

        for committee_slug, group in itertools.groupby(groups, key=outer_grouper):
            
            group = sorted(group, key=inner_grouper)
            
            committee_group = {
                'name': group[0][0],
                'slug': group[0][1],
                'bills': []
            }
            
            
            for bill_slug, bill_group in itertools.groupby(group, key=inner_grouper):
                
                bill_group = list(bill_group)

                bill = {
                    'identifier': bill_group[0][2],
                    'slug': bill_group[0][3],
                    'description': bill_group[0][4],
                    'actions': []
                }
                

                for row in sorted(bill_group, key=lambda x: x[7], reverse=True):
                    action = {
                        'description': row[5],
                        'date': row[6]
                    }
                    
                    bill['actions'].append(action)
                
                committee_group['bills'].append(bill)
                
            committee_updates.append(committee_group)

        return committee_updates

    def find_committee_event_updates(self, committee_ids):
        new_events = ''' 
            SELECT DISTINCT ON (committee.ocd_id, event.ocd_id)
              committee.name,
              committee.slug,
              event.*
            FROM councilmatic_core_event AS event
            JOIN new_event AS new
              ON event.ocd_id = new.ocd_id
            JOIN councilmatic_core_eventparticipant AS p
              ON event.ocd_id = p.event_id
            JOIN councilmatic_core_organization AS committee
              ON p.entity_name = committee.name
            WHERE committee.ocd_id IN %s
            ORDER BY committee.ocd_id,
                     event.ocd_id,
                     event.start_time DESC
        '''

        cursor = connection.cursor()
        cursor.execute(new_events, [tuple(committee_ids)])
        columns = [c[0] for c in cursor.description]

        updates = []
        
        grouper = lambda x: x[1]
        event_groups = sorted(cursor, key=grouper)

        for slug, group in itertools.groupby(event_groups, key=grouper):
            group = list(group)

            committee = {
                'name': group[0][0],
                'slug': group[0][1],
                'events': []
            }
            
            for row in group:
                event = dict(zip(columns[2:], row[2:]))
                committee['events'].append(event)
            
            committee['events'] = sorted(committee['events'], 
                                         key=lambda x: x['start_time'], 
                                         reverse=True)

            updates.append(committee)

        return updates

@django_rq.job
def send_notification_email(user_id=None,
                            user_email=None,
                            bill_action_updates=[],
                            bill_search_updates=[],
                            person_updates=[],
                            committee_action_updates=[],
                            committee_event_updates=[]):
    
    context = {
        # 'user': user, 
        'SITE_META': settings.SITE_META,
        'CITY_VOCAB': settings.CITY_VOCAB,
        'bill_action_updates': bill_action_updates,
        'bill_search_updates': bill_search_updates,
        'person_updates': person_updates,
        'committee_action_updates': committee_action_updates,
        'committee_event_updates': committee_event_updates,
    }

    html = "notifications_email.html"
    txt = "notifications_email.txt"
    html_template = get_template(html)
    text_template = get_template(txt)
    html_content = html_template.render(context)
    text_content = text_template.render(context)
    subject = '{0} Updates!'.format(settings.SITE_META['site_name'])

    msg = EmailMultiAlternatives(subject, 
                                 text_content, 
                                 settings.EMAIL_HOST_USER, 
                                 [user_email])

    msg.attach_alternative(html_content, 'text/html')
    msg.send()



from django.conf.urls import include, url
from django.views.decorators.cache import never_cache

from notifications.views import notifications_login, notifications_logout, \
    notifications_signup, notifications_activation, notifications_account_settings, \
    SubscriptionsManageView, person_subscribe, person_unsubscribe, bill_subscribe, \
    bill_unsubscribe, committee_events_subscribe, committee_events_unsubscribe, \
    committee_actions_subscribe, committee_actions_unsubscribe, search_check_subscription, \
    search_subscribe, search_unsubscribe, events_subscribe, events_unsubscribe, \
    send_notifications

import django_rq


urlpatterns = [
    url(r'^login/$', notifications_login, name='notifications_login'),
    url(r'^logout/$', notifications_logout, name='notifications_logout'),
    url(r'^signup/$', notifications_signup, name='notifications_signup'),
    url(r'^activation/(?P<activation_key>[^/]+)/$', notifications_activation, name='notifications_activation'),
    url(r'^account/settings/$', notifications_account_settings, name='notifications_account_settings'),
    url(r'^account/subscriptions/$', never_cache(SubscriptionsManageView.as_view()), name='subscriptions_manage'),
    # url(r'^notification_loaddata$', notification_loaddata, name='notification_loaddata'),
    # list of things to subscribe/unsubscribe to:
    # - people
    # - committee actions
    # - committee events
    # - bill searches
    # - bill actions
    # - all events
    url(r'^person/(?P<slug>[^/]+)/subscribe/$', person_subscribe, name='person_subscribe'),
    url(r'^person/(?P<slug>[^/]+)/unsubscribe/$', person_unsubscribe, name='person_unsubscribe'),
    url(r'^legislation/(?P<slug>[^/]+)/subscribe/$', bill_subscribe, name='bill_subscribe'),
    url(r'^legislation/(?P<slug>[^/]+)/unsubscribe/$', bill_unsubscribe, name='bill_unsubscribe'),
    url(r'^committee/(?P<slug>[^/]+)/events/subscribe/$',
        committee_events_subscribe, name='committee_events_subscribe'),
    url(r'^committee/(?P<slug>[^/]+)/events/unsubscribe/$',
        committee_events_unsubscribe, name='committee_events_unsubscribe'),
    url(r'^committee/(?P<slug>[^/]+)/actions/subscribe/$',
        committee_actions_subscribe, name='committee_actions_subscribe'),
    url(r'^committee/(?P<slug>[^/]+)/actions/unsubscribe/$',
        committee_actions_unsubscribe, name='committee_actions_unsubscribe'),
    url(r'^search_check_subscription/$', search_check_subscription, name='search_check_subscription'),
    url(r'^search_subscribe/$', search_subscribe, name='search_subscribe'),
    url(r'^search_unsubscribe/$', search_unsubscribe, name='search_unsubscribe'),
    url(r'^events/subscribe/$',
        events_subscribe, name='events_subscribe'),
    url(r'^events/unsubscribe/$',
        events_unsubscribe, name='events_unsubscribe'),
    url(r'^send-notifications/$',
        send_notifications, name='send-notifications'),
    # django-rq: https://github.com/ui/django-rq
    url(r'^django-rq/', include('django_rq.urls')),
]



import json
import urllib

from django import template
from django.db.models.query import QuerySet
from django.utils.safestring import mark_safe

register = template.Library()

# Used in subscriptions_manage.html. XXX: potential (?) site of concern for
# injecting JSON in search facet dicts and re-jsonifying it there (From
# https://stackoverflow.com/questions/4698220/django-template-convert-a-python-list-into-a-javascript-object
# ) Open ticket in Django (with discussion of problematic aspects)
# https://code.djangoproject.com/ticket/17419
@register.filter
def jsonify(object):
    if isinstance(object, QuerySet):
        return mark_safe(serialize('json', object))
    return mark_safe(json.dumps(object))
jsonify.is_safe = True

# Given a search subscription object, successfully reconstruct the
# URL representing it
@register.filter
def custom_reverse_search_url(subscription):
    url = '/search/'
    d = [('q',subscription.search_params['term'])]
    
    for k,vs in subscription.search_params['facets'].items():
        
        for v in vs:
            d.append(("selected_facets","%s_exact:%s" % (k,v)))
    
    url += "?" + urllib.parse.urlencode(d)
    
    return url



from django.template.loader import get_template
from django.conf import settings
from django.core.mail import EmailMultiAlternatives

def send_signup_email(user, host, redirect):
    
    context = {
        'user': user, 
        'SITE_META': settings.SITE_META,
        'host': host,
        'redirect': redirect,
    }

    html = "signup_email.html"
    txt = "signup_email.txt"
    html_template = get_template(html)
    text_template = get_template(txt)
    html_content = html_template.render(context)
    text_content = text_template.render(context)
    subject = '{0} account activation'.format(settings.SITE_META['site_name'])

    # send email confirmation to info@largelots.org
    msg = EmailMultiAlternatives(subject, 
                                 text_content, 
                                 settings.EMAIL_HOST_USER, 
                                 [user.email])

    msg.attach_alternative(html_content, 'text/html')
    msg.send()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-11-21 17:08
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0004_subscriptionprofile'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='billsearchsubscription',
            name='search_facets',
        ),
        migrations.RemoveField(
            model_name='billsearchsubscription',
            name='search_term',
        ),
        migrations.AddField(
            model_name='billsearchsubscription',
            name='search_params',
            field=django.contrib.postgres.fields.jsonb.JSONField(db_index=True, null=True),
        ),
    ]



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-councilmatic-notifications',
    version='0.1.6',
    packages=['notifications'],
    include_package_data=True,
    license='MIT License',  # example license
    description='Core functions for councilmatic.org family',
    long_description=README,
    url='http://councilmatic.org/',
    author='DataMade, LLC',
    author_email='info@datamade.us',
    install_requires=[
        'django-councilmatic>=0.6.9,<0.7',
        'django-rq==0.9.3'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',  # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # Replace these appropriately if you are stuck on Python 2.
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import json
import datetime
import pytz
import sys
import random
import hashlib

from django.shortcuts import render, get_object_or_404
from django.http import HttpResponseRedirect, HttpResponse
from django.conf import settings
from django import forms
from django.utils import timezone

from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMessage
from django.core.cache import cache
from django.core import management

from django.contrib import messages
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.forms import AuthenticationForm, UserCreationForm
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.models import User # XXX TODO: migrate to custom User model https://docs.djangoproject.com/en/1.9/topics/auth/customizing/ http://blog.mathandpencil.com/replacing-django-custom-user-models-in-an-existing-application/ https://www.caktusgroup.com/blog/2013/08/07/migrating-custom-user-model-django/

from django.views.decorators.csrf import csrf_exempt
from django.views.generic import TemplateView, ListView, DetailView

from django.template.loader import get_template
from django.template import Context

from django.db.models import Q
from django.db import IntegrityError

from councilmatic_core.models import Bill, Organization, Person, Event
from notifications.models import PersonSubscription, BillActionSubscription, \
    CommitteeActionSubscription, CommitteeEventSubscription, \
    BillSearchSubscription, EventsSubscription, SubscriptionProfile

from notifications.utils import send_signup_email

app_timezone = pytz.timezone(settings.TIME_ZONE)

class CouncilmaticUserCreationForm(UserCreationForm):
    email = forms.EmailField(label="Email address", required=True,
        help_text="Required.")

    class Meta:
        model = User
        fields = ("username", "email", "password1", "password2")

    def save(self, commit=True):
        user = super().save(commit=False) # this should set the password
        user.email = self.cleaned_data["email"]
        user.is_active = False
        if commit:
            user.save()
        
        profile = SubscriptionProfile()
        profile.user = user
        
        salt = hashlib.sha1(str(random.random()).encode('utf-8')).hexdigest()[:5]
        activation_key = hashlib.sha1((salt + user.username).encode('utf-8')).hexdigest()
        
        profile.activation_key = activation_key
        
        now = timezone.now()
        expiration = now + datetime.timedelta(days=1)
        profile.key_expires = expiration

        profile.save()

        return user

def notifications_activation(request, activation_key):
    
    profile = get_object_or_404(SubscriptionProfile, 
                                activation_key=activation_key)
    
    message_level = 'INFO'
    redirect = reverse('index')
    
    next_url = request.GET.get('next')
    
    if next_url:
        redirect = next_url

    if not profile.user.is_active:
        
        if timezone.now() > profile.key_expires:
            
            salt = hashlib.sha1(str(random.random()).encode('utf-8')).hexdigest()[:5]
            activation_key = hashlib.sha1((salt + profile.user.username).encode('utf-8')).hexdigest()
            
            profile.activation_key = activation_key
            profile.save()
            
            send_signup_email(profile.user, request.get_host())
            
            message = 'Your activation link has expired. A new link has been sent to your email.'
            message_level = 'ERROR'

        else:
            profile.user.is_active = True
            profile.user.save()
            
            redirect = reverse('notifications_login')

            if next_url:
                redirect = '{0}?next={1}'.format(redirect, next_url)

            message = 'Your account has been activated! Login to continue.'

    else:
        message = 'Your account has already been activated.'

    messages.add_message(request, 
                         getattr(messages, message_level), 
                         message)

    return HttpResponseRedirect(redirect)

def notifications_signup(request):
    form = None
    if request.method == 'POST':
        form = CouncilmaticUserCreationForm(data=request.POST)
        if form.is_valid():
            try:
                user = form.save()
                
                redirect = reverse('index')
                if request.GET.get('next'):
                    redirect = request.GET['next']
                
                host = '{0}://{1}'.format(request.scheme, request.get_host())
                send_signup_email(user, host, redirect)
                messages.add_message(request, messages.INFO, 'Check your email for a link to confirm your account!')
                
                return HttpResponseRedirect(redirect)
            except IntegrityError:
                response = HttpResponse('Not able to save form.')
                response.status_code = 500
                return response
        else:
            return render(request, 'notifications_signup.html', {'form': form}, status=500)
    if not form:
        form = CouncilmaticUserCreationForm()
    return render(request, 'notifications_signup.html', {'form': form})

def notifications_login(request):
    form = None
    if request.method == 'POST':
        form = AuthenticationForm(data=request.POST)
        if form.is_valid():
            try:
                user = form.get_user()
                login(request, user)
                
                redirect = reverse('index')
                if request.GET.get('next'):
                    redirect = request.GET['next']

                return HttpResponseRedirect(redirect)
            except IntegrityError:
                response = HttpResponse('Not able to find or login user.')
                response.status_code = 500
                return response
        else:
            return render(request, 'notifications_login.html', {'form': form}, status=500)
    if not form:
        form = AuthenticationForm()
    return render(request, 'notifications_login.html', {'form': form})

def notifications_logout(request):
    logout(request)
    return HttpResponseRedirect('/')

@login_required(login_url='/login/')
def notifications_account_settings(request):
    return HttpResponse('notifications_account_settings')

class SubscriptionsManageView(LoginRequiredMixin, TemplateView):
    template_name = 'subscriptions_manage.html'

    def get_context_data(self, *args, **kwargs):
        context = super(SubscriptionsManageView, self).get_context_data(*args, **kwargs)

        context['person_subscriptions'] = self.request.user.personsubscriptions.all()
        context['committee_action_subscriptions'] = self.request.user.committeeactionsubscriptions.all()
        context['committee_event_subscriptions'] = self.request.user.committeeeventsubscriptions.all()
        context['bill_search_subscriptions'] = self.request.user.billsearchsubscriptions.all()
        context['bill_action_subscriptions'] = self.request.user.billactionsubscriptions.all()
        context['events_subscriptions'] = self.request.user.eventssubscriptions.all()

        return context

@login_required(login_url='/login/')
def bill_subscribe(request, slug):
    bill = Bill.objects.get(slug=slug)
    (bill_action_subscription, created) = BillActionSubscription.objects.get_or_create(user=request.user, bill=bill)

    cache.delete('subscriptions_manage')

    return HttpResponse('Subscribed to bill %s.' % str(bill))

@login_required(login_url='/login/')
def bill_unsubscribe(request, slug):
    bill = Bill.objects.get(slug=slug)

    try:
        bill_action_subscription = BillActionSubscription.objects.get(user=request.user, bill=bill)
    except ObjectDoesNotExist:
        response = HttpResponse('This bill subscription does not exist.')
        response.status_code = 500
        return response

    bill_action_subscription.delete()

    return HttpResponse('Unsubscribed from bill %s.' % str(bill))

@login_required(login_url='/login/')
def person_subscribe(request, slug):
    person = Person.objects.get(slug=slug)

    (person_subscription, created) = PersonSubscription.objects.get_or_create(user=request.user, person=person)

    return HttpResponse('Subscribed to person %s.' % str(person))

@login_required(login_url='/login/')
def person_unsubscribe(request, slug):
    person = Person.objects.get(slug=slug)

    try:
        person_subscription = PersonSubscription.objects.get(user=request.user, person=person)
    except ObjectDoesNotExist:
        response = HttpResponse('This person subscription does not exist.')
        response.status_code = 500
        return response

    person_subscription.delete()

    return HttpResponse('Unsubscribed from person %s.' % str(person))

@login_required(login_url='/login/')
def committee_events_subscribe(request, slug):
    committee = Organization.objects.get(slug=slug)
    (committee_events_subscription, created) = CommitteeEventSubscription.objects.get_or_create(user=request.user, committee=committee)

    return HttpResponse('Subscribed to events of %s.' % str(committee))

@login_required(login_url='/login/')
def committee_events_unsubscribe(request, slug):
    committee = Organization.objects.get(slug=slug)
    try:
        committee_events_subscription = CommitteeEventSubscription.objects.get(user=request.user, committee=committee)
    except ObjectDoesNotExist:
        response = HttpResponse('This committee event subscription does not exist.')
        response.status_code = 500
        return response

    committee_events_subscription.delete()
    return HttpResponse('Unsubscribed from events of %s.' % str(committee))

@login_required(login_url='/login/')
def committee_actions_subscribe(request, slug):
    committee = Organization.objects.get(slug=slug)
    (committee_actions_subscription, created) = CommitteeActionSubscription.objects.get_or_create(user=request.user, committee=committee)

    return HttpResponse('Subscribed to actions of %s.' % str(committee))

@login_required(login_url='/login/')
def committee_actions_unsubscribe(request, slug):
    committee = Organization.objects.get(slug=slug)

    try:
        committee_actions_subscription = CommitteeActionSubscription.objects.get(user=request.user, committee=committee)
    except ObjectDoesNotExist:
        response = HttpResponse('This committee action subscription does not exist.')
        response.status_code = 500
        return response

    committee_actions_subscription.delete()

    return HttpResponse('Unsubscribed from actions of %s.' % str(committee))

@login_required(login_url='/login/')
def search_subscribe(request):
    q = request.POST.get('query')
    selected_facets = request.POST.get('selected_facets')
    search_params = {'term': q, 'facets': json.loads(selected_facets)}
    (bss, created) = BillSearchSubscription.objects.get_or_create(user=request.user, 
                                                                  search_params=search_params)

    return HttpResponse('Subscribed to search for: %s.' % q)

@login_required(login_url='/login/')
def search_unsubscribe(request):
    q = request.POST.get('query')
    selected_facets = request.POST.get('selected_facets')
    selected_facets_json = json.loads(selected_facets)
    search_params = {'term': q, 'facets': selected_facets_json}

    try:
        bss = BillSearchSubscription.objects.get(user=request.user,
                                                 search_params__exact=search_params)
    except ObjectDoesNotExist:
        response = HttpResponse('This search subscription does not exist.')
        response.status_code = 500
        return response

    bss.delete()
    return HttpResponse('Unsubscribed from search for: %s.' % q)

@login_required(login_url='/login/')
def search_check_subscription(request):
    q = request.POST.get('query')
    selected_facets = request.POST.get('selected_facets')
    selected_facets_json = json.loads(selected_facets)
    search_params = {'term': q, 'facets': selected_facets_json}
    
    try:
        bss = BillSearchSubscription.objects.get(user=request.user, 
                                                 search_params__exact=search_params)

    except ObjectDoesNotExist:
        response = HttpResponse('This bill search subscription does not exist.')
        response.status_code = 500
        return response

    return HttpResponse('true')

# TODO: Make these do something
@login_required(login_url='/login/')
def events_subscribe(request):
    (events_subscription, created) = EventsSubscription.objects.get_or_create(user=request.user)

    return HttpResponse('%s unsubscribed from all events.' % request.user)

@login_required(login_url='/login/')
def events_unsubscribe(request):
    try:
        events_subscription = EventsSubscription.objects.get(user=request.user)
    except ObjectDoesNotExist:
        response = HttpResponse('This event subscription does not exist.')
        response.status_code = 500
        return response

    events_subscription.delete()

    return HttpResponse('%s unsubscribed from all events.' % str(request.user))

@login_required(login_url='/login/')
def send_notifications(request):
    
    management.call_command('send_notifications', users=request.user.username)

    return HttpResponse(json.dumps({'status': 'ok'}), content_type='application/json')
# The function worker_handle_notification_email() is invoked when the 'notifications_emails' queue (notification_emails_queue)
# is woken up.


